# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Low-level communication with the Transmission daemon"""

from ...logging import make_logger
log = make_logger(__name__)

import asyncio
import aiohttp
import json
import textwrap
from blinker import Signal
from urllib.parse import urlsplit
import re
import warnings

from ..errors import (URLParserError, ConnectionError, RPCError, AuthError, ClientError)


AUTH_ERROR_CODE = 401
CSRF_ERROR_CODE = 409
CSRF_HEADER = 'X-Transmission-Session-Id'
TIMEOUT = 10


class TransmissionURL():
    _DEFAULT = urlsplit('http://localhost:9091/transmission/rpc')
    _ATTRS = ('scheme', 'netloc', 'path', 'query', 'fragment', 'username',
              'password', 'hostname', 'port')

    _SCHEME_REGEX = re.compile('^\w+://')

    def __init__(self, url='http://localhost:9091/transmission/rpc'):
        url = str(url)
        # Insert default scheme before parsing the URL. Otherwise, urlsplit
        # is confused.
        if not self._SCHEME_REGEX.match(url):
            url = self._DEFAULT.scheme + '://' + url

        # urlsplit doesn't seem to raise meaningful errors,
        # e.g. 'localhost:123:456' raises:
        # ValueError: invalid literal for int() with base 10: '123:456'
        try:
            parsed_url = urlsplit(url, allow_fragments=False)
            urldict = {}
            for attr in self._ATTRS:
                urldict[attr] = getattr(parsed_url, attr)
        except Exception:
            raise URLParserError(url)

        # Some more defaults
        if urldict['port'] is None:
            urldict['port'] = self._DEFAULT.port
        if urldict['path'] in ('/', ''):
            urldict['path'] = self._DEFAULT.path
        self.url = urldict

    def str(self, path=False, scheme=False, port=False, password=False):
        if None not in (self.url['username'], self.url['password']):
            if password:
                urlfmt = '{username}:{password}@{hostname}'
            else:
                urlfmt = '{username}@{hostname}'
        else:
            urlfmt = '{hostname}'
        if scheme:
            urlfmt = '{scheme}://' + urlfmt
        if port:
            urlfmt = urlfmt + ':{port}'
        if path:
            urlfmt = urlfmt + '{path}'
        return urlfmt.format(**self.url)

    def __repr__(self):
        return self.str(path=True, scheme=True, port=True, password=True)

    def __str__(self):
        scheme = self.url['scheme'] != self._DEFAULT.scheme
        path   = self.url['path']   != self._DEFAULT.path
        port   = self.url['port']   != self._DEFAULT.port
        return self.str(path=path, scheme=scheme, port=port, password=False)

    def __getattr__(self, attr):
        try:
            return self.url[attr]
        except KeyError:
            raise AttributeError(attr)

    def __eq__(self, other):
        return str(self) == str(other)

    def __ne__(self, other):
        return not self.__eq__(other)


class TransmissionRPC():
    """Low-level AsyncIO Transmission RPC communication

    This class handles connecting to a Transmission daemon via the RPC
    interface.  It does not implement the RPC protocol, only basic things like
    authentication, sending requests and receiving responses.  High-level RPC
    are done in the *API classes.
    """

    def __init__(self, url='localhost:9091', loop=None):
        # Use double underscores because TransmissionAPI inherits from this
        # class; this way, we don't have to worry about name collisions.
        self.loop = loop if loop is not None else asyncio.get_event_loop()
        self.__url = TransmissionURL(url)
        self.__headers = {'content-type': 'application/json'}
        self.__session = None
        self.__connect_exception = None
        self.__connecting_lock = asyncio.Lock(loop=loop)
        self.__connection_tested = False
        self.__timeout = TIMEOUT
        self.__on_connected = Signal()
        self.__on_disconnected = Signal()
        self.__on_error = Signal()
        self.__version = None
        self.__rpcversion = None
        self.__rpcversionmin = None

    def __del__(self, _warnings=warnings):
        if self.__session is not None and not self.__session.closed:
            _warnings.warn('disconnect() wasn\'t called', ResourceWarning)
            self.__session.close()

    def on(self, signal, callback, autoremove=True):
        """Register `callback` for `signal`

        signal: 'connected', 'disconnected' or 'error'
        callback: a callable that receives the RPC URL and, for 'error', the
                  exception

        Callbacks are automatically unsubscribed when they are
        garbage-collected.
        """
        try:
            # Attributes with '__' become '_Classname__attribute'
            sig = getattr(self, '_TransmissionRPC__on_' + signal)
        except AttributeError:
            raise ValueError('Unknown signal: {!r}'.format(signal))
        else:
            if not isinstance(sig, Signal):
                raise ValueError('Unknown signal: {!r}'.format(signal))
            else:
                log.debug('Registering %r for %r event', callback, signal)
                sig.connect(callback, weak=autoremove)

    @property
    def version(self):
        """Version of the Transmission daemon or None if not connected"""
        return self.__version

    @property
    def rpcversion(self):
        """RPC version of the Transmission daemon or None if not connected"""
        return self.__rpcversion

    @property
    def rpcversionmin(self):
        """Oldest RPC version supported by Transmission daemon or None if not connected"""
        return self.__rpcversionmin

    @property
    def url(self):
        """Transmission's RPC URL without the path"""
        return self.__url

    @url.setter
    def url(self, url):
        self.__url = TransmissionURL(url)

    @property
    def timeout(self):
        """Number of seconds to try to connect before giving up"""
        return self.__timeout

    @timeout.setter
    def timeout(self, timeout):
        self.__timeout = timeout

    @property
    def connected(self):
        """Return True if connected, False otherwise"""
        return self.__session is not None and not self.__session.closed \
            and self.__connection_tested

    async def connect(self, url=None, timeout=None):
        """Connect to running daemon

        url: URL to the Transmission RPC interface or None for default
             (see TransmissionURL)
        timeout: Maximum number of seconds before attempt to connect fails

        Raises RPCError, ConnectionError or AuthError.
        """
        if self.__connecting_lock.locked():
            # Someone else is currently connecting.  Wait for them to finish.
            log.debug('Waiting for other connection to establish')
            await self.__connecting_lock.acquire()
            log.debug('Other connect() call finished')
            self.__connecting_lock.release()

            # Check if connection has the url we want
            if url is not None and self.url != TransmissionURL(url):
                log.debug('Reconnecting because %s != %s', self.__url, url)
                await self.connect(url=url, timeout=timeout)

            # The other connect() call croaked for some reason, and our caller
            # expects the same exception.  Calling connect() again should
            # raise the same error, but that's too much recursion for my tiny,
            # little mind.
            elif self.__connect_exception is not None:
                raise self.__connect_exception

            # Looks like we're connected.  Our intended timeout may differ
            # from what the previous call set, but there's no need to
            # re-connect.
            elif timeout is not None and self.timeout != timeout:
                self.timeout = timeout

        else:
            async with self.__connecting_lock:
                log.debug('Acquired connect() lock')
                if timeout is not None:
                    self.timeout = timeout

                # Reconnect if URL is specified
                if url is not None:
                    if self.connected:
                        log.debug('Reconnecting to %s', url)
                        await self.disconnect('reconnecting to %s' % url)
                    self.__url = TransmissionURL(url)

                # If we're not connected (because new URL or we weren't
                # connected in the first place), create new session.
                if not self.connected:
                    log.debug('Connecting to %s (timeout=%ss)', self.url, self.timeout)
                    session_args = {'loop': self.loop}
                    if self.__url.username or self.__url.password:
                        session_args['auth'] = aiohttp.BasicAuth(self.__url.username,
                                                                 self.__url.password)
                    self.__session = aiohttp.ClientSession(**session_args)
                    skip_connected_callback = False
                else:
                    log.debug('Already connected to %s', self.url)
                    skip_connected_callback = True

                # Check if connection works.  If we were already connected, we
                # still want to check if the old connection still works, but
                # we don't want to report the working connection again to
                # callbacks (skip_connected_callback).
                log.debug('Initializing new connection to %s', self.url)
                try:
                    test_request = json.dumps({'method':'session-get'})
                    info = await self.__send_request(test_request)
                except ClientError as e:
                    log.debug('Caught %s during initialization: %r', type(e).__name__, e)
                    self.__connect_exception = e
                    await self.disconnect('Connection failed during initialization')
                    self.__on_error.send(self.__url, error=e)
                    raise
                else:
                    self.__version = info['version']
                    self.__rpcversion = info['rpc-version']
                    self.__rpcversionmin = info['rpc-version-minimum']
                    self.__connection_tested = True
                    self.__connect_exception = None
                    log.debug('Connection established: %s', self.url)
                    if not skip_connected_callback:
                        self.__on_connected.send(self.__url)

                log.debug('Releasing connect() lock')

    async def disconnect(self, reason=None):
        """Disconnect if connected

        reason: Why are we disconnecting? Only used in a debugging message.
        """
        await self.__reset()
        log.debug('Disconnecting from %s (%s)',
                  self.url, reason if reason is not None else 'for no reason')
        self.__on_disconnected.send(self.__url)

    async def __reset(self):
        if self.__session is not None:
            await self.__session.close()
        self.__session = None
        self.__version = None
        self.__rpcversion = None
        self.__rpcversionmin = None
        self.__connection_tested = False

    async def __post(self, data):
        with aiohttp.Timeout(self.timeout, loop=self.loop):
            try:
                response = await self.__session.post(repr(self.__url), data=data, headers=self.__headers)
            except aiohttp.ClientResponseError as e:
                log.debug('Caught ClientResponseError during POST request')
                raise ConnectionError(str(self.url))
            else:
                if response.status == CSRF_ERROR_CODE:
                    # Send request again with CSRF header
                    self.__headers[CSRF_HEADER] = response.headers[CSRF_HEADER]
                    log.debug('Setting CSRF header: %s = %s',
                              CSRF_HEADER, response.headers[CSRF_HEADER])
                    await response.release()
                    return await self.__post(data)

                elif response.status == AUTH_ERROR_CODE:
                    await response.release()
                    log.debug('Authentication failed')
                    raise AuthError(str(self.url))

                else:
                    try:
                        answer = await response.json()
                    except ValueError as e:
                        text = textwrap.shorten(await response.text(),
                                                100, placeholder='...')
                        raise RPCError('Server sent malformed JSON: {}'.format(text))
                    else:
                        return answer

    async def __send_request(self, post_data):
        """Send RPC POST request to daemon

        post_data: Any valid RPC request as JSON string

        If applicable, returns response['arguments']['torrents'] or
        response['arguments'], otherwise response.

        Raises ClientError.
        """
        try:
            log.debug('Sending POST request: %r', post_data)
            answer = await self.__post(post_data)
        except OSError as e:
            log.debug('Caught OSError: %r', e)
            raise ConnectionError(str(self.url))
        except asyncio.TimeoutError as e:
            log.debug('Caught TimeoutError: %r', e)
            raise ConnectionError('Timeout after {}s: {}'.format(self.timeout, self.url))
        else:
            if answer['result'] != 'success':
                raise RPCError(answer['result'].capitalize())
            else:
                if 'arguments' in answer:
                    if 'torrents' in answer['arguments']:
                        return answer['arguments']['torrents']
                    else:
                        return answer['arguments']
                return answer

    def __getattr__(self, method):
        """Return asyncio coroutine that sends RPC request and returns response

        method: Any method from the RPC specs with every '-' replaced with '_'.
                For arguments see the RPC specs.

        Example:
        >>> stats = await client.session_stats()
        >>> torrents = await client.torrent_get(ids=(1,2,3), fields=('status','name'))

        Raises RPCError, ConnectionError, AuthError
        """
        realmethod = method.replace('_', '-')
        log.debug('Generating RPC request method: %r', realmethod)

        async def request(arguments={}, autoconnect=True, **kwargs):
            if not self.connected:
                if autoconnect:
                    log.debug('Autoconnecting')
                    await self.connect()
                else:
                    log.debug('Not connected and autoconnect=%r - %r returns None',
                              autoconnect, method)
                    return None

            arguments.update(**kwargs)
            rpc_request = json.dumps( {'method':realmethod, 'arguments':arguments} )
            try:
                return await self.__send_request(rpc_request)
            except ClientError as e:
                log.debug('Caught ClientError in %r request: %r', method, e)

                # RPCError does not mean host is unreachable, there was just a
                # misunderstanding, so we're still connected.
                if not isinstance(e, RPCError):
                    await self.disconnect(str(e))

                self.__on_error.send(self.__url, error=e)
                raise

        request.__name__ = method
        request.__qualname__ = method
        return request



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

try:
    import GeoIP

except ImportError:
    GEOIP_AVAILABLE = False
    def country_code(addr):
        return '?'

else:
    GEOIP_AVAILABLE = True
    _geoip = GeoIP.new(GeoIP.GEOIP_MEMORY_CACHE)
    def country_code(addr):
        return _geoip.country_code_by_addr(addr) or '?'



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import urwid
import collections


class Style():
    """Map standard attributes to those defined in a urwid palette

    prefix: common prefix of all attributes
    extras: additional attributes (e.g. 'header')
    modes: additional subsections with 'focused' and 'unfocused' attributes
           (e.g. 'highlighted')
    focusable: True if 'focused' attributes should be mapped, False otherwise
    """

    def __init__(self, prefix, modes=(), extras=(), focusable=True):
        self._attribs = {
            'unfocused': self.dotify(prefix, 'unfocused')
        }
        for extra in extras:
            self._attribs[extra] = self.dotify(prefix, extra)

        if focusable:
            self._attribs['focused'] = self.dotify(prefix, 'focused')

        for mode in modes:
            self._attribs[mode+'.focused'] = self.dotify(prefix, mode, 'focused')
            self._attribs[mode+'.unfocused'] = self.dotify(prefix, mode, 'unfocused')

    def attrs(self, mode=None, focused=False):
        """Get attributes as specified in the urwid palette

        mode: one of the modes specified during initialization
        focused: If True the '...focused' attributes are returned,
                 '...unfocused otherwise
        """
        mode = '' if not mode else mode
        name = self.dotify(mode, 'focused' if focused else 'unfocused')
        if name in self._attribs:
            return self._attribs[name]
        else:
            return self._attribs[mode]

    @property
    def focus_map(self):
        """Map of all '...unfocused' -> '...focused' attributes"""
        focus_map = {}
        attribs = self._attribs
        for name in attribs:
            if name.endswith('unfocused'):
                name_focused = name[:-9] + 'focused'
                focus_map[attribs[name]] = attribs[name_focused]
        return focus_map

    @staticmethod
    def dotify(*strings):
        """Join non-empty strings with a '.'"""
        return '.'.join(x for x in (strings) if x)


# TODO: It would be great if some columns (e.g. rate-up/down) would shrink to
#       the widest visible value (including its header).  Other columns would
#       share the remaining width using ('weight', x), giving some of them
#       (e.g. name) more space than others (e.g. path).
class CellWidgetBase(urwid.WidgetWrap):
    style = collections.defaultdict(lambda: 'default')
    header = urwid.Padding(urwid.Text('NO HEADER SPECIFIED'))
    width = ('weight', 100)
    align = 'right'

    def __init__(self):
        self.value = None
        self.text = urwid.Text('', wrap='clip', align=self.align)
        self.attrmap = urwid.AttrMap(self.text, self.style.attrs('unfocused'))
        return super().__init__(self.attrmap)

    def update(self, data):
        self.data.update(data)
        new_value = self.get_value()
        if self.value != new_value:
            self.value = new_value
            self.text.set_text(str(new_value))
            attr = self.style.attrs(self.get_mode(), focused=False)
            self.attrmap.set_attr_map({None: attr})

    def get_value(self):
        raise NotImplementedError()

    def get_mode(self):
        return None


from . import hooks



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...logging import make_logger
log = make_logger(__name__)

import blinker
from collections import abc
import os

from ..poll import RequestPoller
from .. import convert
from .. import constants as const
from functools import wraps


def setting(method):
    """Decorator for SettingsAPI properties"""
    @wraps(method)
    def wrapped(self):
        if self._raw is None:
            return const.DISCONNECTED
        else:
            return method(self)
    return property(wrapped)


class SettingsAPI(abc.Mapping):
    """Transmission daemon settings

    get_* methods are coroutine functions and fetch values from the server.

    Cached values that are updated every `interval` seconds are available as
    properties with 'get_' removed from the equivalent method name.  For
    example, `get_path_incomplete` as an asyncronous method, `path_incomplete`
    is a syncronous property.

    Cached values are also available as items, e.g. settings['path-incomplete'].
    (Remove 'get_' from the method name and replace '_' with '-'.)

    To update cached values, use `update` (async) or `poll` (sync).
    """

    # Forward some methods and properties to the internal poller
    def __getattr__(self, attr):
        if attr in ('start', 'stop', 'poll', 'running'):
            return getattr(self._poller, attr)
        raise AttributeError('{!r} object has no attribute {!r}'
                             .format(type(self).__name__, attr))

    @property
    def interval(self):
        return self._poller.interval

    @interval.setter
    def interval(self, interval):
        self._poller.interval = interval


    # Mandatory abc.Mapping methods
    def __iter__(self):
        for attr in dir(self):
            if attr.startswith('get_'):
                yield attr[4:].replace('_', '-')

    def __len__(self):
        return len(tuple(iter(self)))


    def __init__(self, srvapi, interval=1):
        self._raw = None      # Raw dict from 'session-get' or None if not connected
        self._cache = {}      # Cached values convert by _AFTER_GET
        self._srvapi = srvapi
        self._get_timestamp = 0
        self._on_update = blinker.Signal()

        # autoconnect must be True so the CLI 'help' command can display
        # current values (e.g. 'help srv.limit.rate.down').
        self._poller = RequestPoller(self._srvapi.rpc.session_get,
                                     autoconnect=True,
                                     interval=interval,
                                     loop=srvapi.loop)
        self._poller.on_response(self._handle_session_get)
        self._poller.on_error(self._handle_error)

    def _handle_error(self, e):
        log.debug('Ignoring {!r}'.format(e))

    def clearcache(self):
        """Clear cached settings"""
        self._raw = None
        self._cache = {}

    def on_update(self, callback, autoremove=True):
        """Register `callback` to be called when settings have changed

        `callback` gets the instance of this class.

        If `autoremove` is True, `callback` is removed automatically when it
        is deleted.
        """
        log.debug('Registering %r to receive settings updates', callback)
        self._on_update.connect(callback, weak=autoremove)

    def _handle_session_get(self, response):
        """Request update from server"""
        log.debug('Handling periodic settings update')
        self._raw = response
        self._cache = {}
        self._on_update.send(self)

    async def update(self):
        """Request update from server"""
        log.debug('Requesting settings update')
        self._handle_session_get(await self._poller.request())

    def __getitem__(self, key):
        return getattr(self, key.replace('-', '_'))

    async def _set(self, dct):
        """Set dictionary `settings` with the 'session-set' RPC method"""
        log.debug('Setting %s', dct)
        await self._srvapi.rpc.session_set(dct)
        await self.update()


    # Rate limits

    def _rate_limit_keys(self, direction, alt):
        if alt:
            key_value = 'alt-speed-'+direction
            key_enabled = 'alt-speed-enabled'
        else:
            key_value = 'speed-limit-'+direction
            key_enabled = 'speed-limit-'+direction+'-enabled'
        return (key_value, key_enabled)

    def _get_rate_limit(self, direction, alt=False):
        key_value, key_enabled = self._rate_limit_keys(direction, alt)
        if self._raw[key_enabled]:
            # Transmission reports kilobytes
            return convert.bandwidth(self._raw[key_value]*1000, unit='byte')
        else:
            return const.UNLIMITED

    def _set_rate_limit(self, limit, direction, alt=False):
        key_value, key_enabled = self._rate_limit_keys(direction, alt)
        if limit in (const.ENABLED, True):
            return {key_enabled: True}
        elif limit in (const.UNLIMITED, const.DISABLED, False) or \
             isinstance(limit, (int, float)) and limit < 0:
            return {key_enabled: False}
        else:
            l = convert.bandwidth(limit)
            if l.unit == 'b':
                l /= 8  # Convert to kilobytes
            return {key_value: int(l/1000), key_enabled: True}


    @setting
    def rate_limit_up(self):
        """Cached upload rate limit

        Returns a Number object created by the `bandwidth` converter in the
        `convert` module or one of the constants: UNLIMITED, DISCONNECTED
        """
        if 'rate_limit_up' not in self._cache:
            self._cache['rate_limit_up'] = self._get_rate_limit('up', alt=False)
        return self._cache['rate_limit_up']

    async def get_rate_limit_up(self):
        """Refresh cache and return `rate_limit_up`"""
        await self.update()
        return self.rate_limit_up

    async def set_rate_limit_up(self, limit):
        """Set upload rate limit to `limit`

        The `bandwidth` converter in the `convert` module is used to determine
        the unit (bits or bytes) of `limit`.

        An existing limit is disabled if `limit` is False, one of the
        constants UNLIMITED or DISABLED, or a number < 0.

        An existing limit is enabled if `limit` is True or the constant
        ENABLED.
        """
        await self._set(self._set_rate_limit(limit, 'up', alt=False))


    @setting
    def rate_limit_down(self):
        """Cached download rate limit (see `rate_limit_up`)"""
        if 'rate_limit_down' not in self._cache:
            self._cache['rate_limit_down'] = self._get_rate_limit('down', alt=False)
        return self._cache['rate_limit_down']

    async def get_rate_limit_down(self):
        """Refresh cache and return `rate_limit_down`"""
        await self.update()
        return self.rate_limit_down

    async def set_rate_limit_down(self, limit):
        """Set upload rate limit to `limit` (see `set_rate_limit_up`)"""
        await self._set(self._set_rate_limit(limit, 'down', alt=False))


    @setting
    def alt_rate_limit_up(self):
        """Cached alternative upload rate limit (see `rate_limit_up`)"""
        if 'rate_limit_up' not in self._cache:
            self._cache['rate_limit_up'] = self._get_rate_limit('up', alt=True)
        return self._cache['rate_limit_up']

    async def get_alt_rate_limit_up(self):
        """Refresh cache and return `alt_rate_limit_up`"""
        await self.update()
        return self.alt_rate_limit_up

    async def set_alt_rate_limit_up(self, limit):
        """Set alternative upload rate limit to `limit` (see `set_rate_limit_up`)"""
        await self._set(self._set_rate_limit(limit, 'up', alt=True))


    @setting
    def alt_rate_limit_down(self):
        """Cached alternative download rate limit (see `rate_limit_up`)"""
        if 'rate_limit_down' not in self._cache:
            self._cache['rate_limit_down'] = self._get_rate_limit('down', alt=True)
        return self._cache['rate_limit_down']

    async def get_alt_rate_limit_down(self):
        """Refresh cache and return `alt_rate_limit_down`"""
        await self.update()
        return self.alt_rate_limit_down

    async def set_alt_rate_limit_down(self, limit):
        """Set alternative upload rate limit to `limit` (see `set_rate_limit_up`)"""
        await self._set(self._set_rate_limit(limit, 'down', alt=True))


    # Paths

    def _absolute_path(self, path, cwd):
        return os.path.normpath(os.path.join(cwd, path))

    @setting
    def path_complete(self):
        """Path to directory where torrent files are put"""
        return os.path.normpath(self._raw['download-dir'])

    async def get_path_complete(self):
        """Get path to directory where torrent files are put"""
        await self.update()
        return self.path_complete

    async def set_path_complete(self, path):
        """Set path to directory where torrent files are put

        If path is relative (i.e. doesn't start with '/'), it is relative to
        `get_path_complete`.
        """
        path = self._absolute_path(path, await self.get_path_complete())
        await self._set({'download-dir': path})

    @setting
    def path_incomplete(self):
        """Path to directory where incomplete torrent files are put"""
        if not self._raw['incomplete-dir-enabled']:
            return const.DISABLED
        else:
            return os.path.normpath(self._raw['incomplete-dir'])

    async def get_path_incomplete(self):
        """Get path to directory where incomplete torrent files are put

        Returns `constants.DISABLED` if incomplete files are stored together
        with complete ones.
        """
        await self.update()
        return self.path_incomplete

    async def set_path_incomplete(self, path):
        """Set path to directory where torrent files are put before they are complete

        If `path` is not a `str` instance, it is evaluated as a bool and this
        feature is enabled or disabled without changing the path.
        """
        if isinstance(path, str):
            await self.update()
            current_path = self._raw['incomplete-dir']
            await self._set({'incomplete-dir': self._absolute_path(path, current_path),
                             'incomplete-dir-enabled': True})
        else:
            await self._set({'incomplete-dir-enabled': bool(path)})


    # Other settings

    @setting
    def dht(self):
        """Whether DHT is used to discover peers"""
        return const.ENABLED if self._raw['dht-enabled'] else const.DISABLED

    async def get_dht(self):
        """Refresh cache and return `dht`"""
        await self.update()
        return self.dht

    async def set_dht(self, enabled):
        """Whether DHT should be used to disover peers"""
        await self._set({'dht-enabled': bool(enabled)})


    @setting
    def lpd(self):
        """Whether Local Peer Discovery is used to discover peers"""
        return const.ENABLED if self._raw['lpd-enabled'] else const.DISABLED

    async def get_lpd(self):
        """Refresh cache and return `lpd`"""
        await self.update()
        return self.lpd

    async def set_lpd(self, enabled):
        """Whether Local Peer Discovery should be used to disover peers"""
        await self._set({'lpd-enabled': bool(enabled)})


    @setting
    def encryption(self):
        """Whether protocol encryption is used to mask BitTorrent traffic

        One of the strings 'required', 'preferred' or 'tolerated'.
        """
        return self._raw['encryption']

    async def get_encryption(self):
        """Refresh cache and return `encryption`"""
        await self.update()
        return self.encryption

    async def set_encryption(self, encryption):
        """Whether encryption is 'required', 'preferred' or 'tolerated'"""
        if encryption in ('required', 'preferred', 'tolerated'):
            await self._set({'encryption': encryption})
        else:
            raise ValueError("Must be one of 'required', 'preferred', 'tolerated', not {!r}"
                             .format(encryption))
